import pandas as pd


def create_calendar_table(start, end, columns=['day', 'month', 'year', 'week', 'weekday', 'month_name', 'weekday_name'],
                          locale=None):
    calendar = pd.DataFrame(pd.date_range(
        start=start, end=end), columns=['date'])
    if 'day' in columns:
        calendar['day'] = calendar.date.dt.day
    if 'month' in columns:
        calendar['month'] = calendar.date.dt.month
    if 'year' in columns:
        calendar['year'] = calendar.date.dt.year
    if 'week' in columns:
        calendar['week'] = calendar.date.dt.week
    if 'weekday' in columns:
        calendar['weekday'] = calendar.date.dt.weekday
    if 'month_name' in columns:
        calendar['month_name'] = calendar.date.dt.month_name()
    if 'weekday_name' in columns:
        calendar['weekday_name'] = calendar.date.dt.day_name()
    return calendar



from datetime import datetime as dt

formato = "%d/%m/%y %H:%M:%S"


class Timer():
    """
    Timer class
    Declare a timer object that measure the time 
    of a running application
    """

    def __init__(self, start_now=False):
        self.start_time = None
        self.stop_time = None
        self.duration = None
        if start_now:
            self.start()

    def start_time_on_format(self, format=formato):
        return self.start_time.strftime(formato)

    def stop_time_on_format(self, format=formato):
        if self.duration is None:
            return "not stoped"
        return self.stop_time.strftime(formato)

    def __repr__(self):
        return f'\n started: {str(self.start_time)} \n finished:{str(self.stop_time)} \n duration:{str(self.duration)}'

    def start(self):
        if self.start_time is None:
            self.start_time = dt.datetime.now()
        else:
            print('Started at: {}'.format(self.start_time))

    def stop(self):
        if self.stop_time is None:
            self.stop_time = dt.datetime.now()
            self.duration = self.stop_time - self.start_time
        else:
            print('Stopped at: {}'.format(self.stop_time))


class Timers():
    """    
    Multi Timer class
    Dictionary with multiple timers to divide steps in a aplication
    and can be called without declaring keys
    """

    def __init__(self, start_now=False, n=1, timers_names=None):
        self.timers = dict()
        self.__starts = 0
        self.__stops = 0

    def __repr__(self):
        return f'{self.timers}'

    def create_timer(self, timer_name):
        self.timers[timer_name] = Timer()

    def start(self, timer_name=None):
        if timer_name is None:
            timer_name = self.__starts
            self.__starts += 1

        self.create_timer(timer_name)
        self.timers[timer_name].start()

    def stop(self, timer_name=None):
        if timer_name is None:
            timer_name = self.__stops
            self.timers[timer_name].stop()
            self.__stops += 1
            return
        else:
            self.timers[timer_name].stop()



from jh_utils.data.sql.connection import create_connection, create_string_connection
from jh_utils.data.sql.manipulate_db import (
    get_tables,
    drop_table,
    delete_table,
    get_schemas,
    create_schema,
    drop_schema,
)
from dotenv import dotenv_values
from sqlalchemy import inspect

doc = """
    env: db, user, pass, host, port
    ----------------------------
    example .env
    host=weather
    host1=weather
    db=weather
    user=weather
    pass=weather12
    port=5400
    schema=raw
    """


class DB:
    __doc__ = doc

    def __init__(self, db, user, password, host, port):
        self.db = db
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def engine(self):
        return create_connection(
            database=self.db,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
        )

    def uri(self):
        return create_string_connection(
            database=self.db,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
        )

    def __repr__(self) -> str:
        return f"""host: {self.host}\ndb:{self.db}"""

    # ! table
    def drop_table(self, table, schema):
        return drop_table(table, schema=schema, engine=self.engine())

    def delete_table(self, table_name, schema):
        delete_table(table_name, schema, self.engine(), close_connection=True)

    def get_schemas(self):
        return get_schemas(self.engine())

    def get_tables(self, schema):
        return get_tables(schema, self.engine())

    # ! schema
    def drop_schema(self, schema):
        drop_schema(schema=schema, engine=self.engine())

    def create_schema(self, schema_nema):
        create_schema(schema_nema, self.engine())

    def run_sql(self, sql):
        conn = self.engine().connect()
        conn.execute(sql)


##
# ? Second form to create the object
##


def create_object_DB(env_dict):
    __docstring__ = doc
    db_object = DB(
        db=env_dict['db'],
        user=env_dict['user'],
        password=env_dict['pass'],
        host=env_dict['host'],
        port=env_dict['port'],
    )
    return db_object


def create_object_DB_by_envfile(path='.env'):
    __docstring__ = doc
    env = dotenv_values(path)
    db_obect_created = create_object_DB(env)
    return db_obect_created



import os


def bash(command: str):
    os.system(command)


def create(name, is_folder=False):
    if is_folder == True:
        os.mkdir(name)
        return
    os.mkfifo(name)


def remove(name, is_folder=False):
    if is_folder == True:
        os.rmdir(name)
        return
    os.remove(name)


def ls(path: str = '.', contains=[''], not_contains=[''], files=True, directories=True):
    # just to transform string in list if necessary
    if type(contains) == str:
        contains = [contains]
    if type(contains) == str:
        not_contains = [not_contains]

    # set first list
    if files == True and directories == True:
        final_ls = os.listdir(path)
    elif files == True:
        final_ls = list(filter(os.path.isfile, os.listdir(path)))
    elif directories == True:
        final_ls = list(filter(os.path.isdir, os.listdir(path)))
    else:
        return 'ERROR: DIRECTORIES OF FILES NEED TO BE TRUE'

    # filter list
    for i in contains:
        final_ls = list(filter(lambda k: i in k, final_ls))
    for i in not_contains:
        final_ls = list(filter(lambda k: i in k, final_ls))

    return final_ls



import os
import json
from numpy.random import randint


def open_json(json_file):
    with open(json_file) as data:
        return json.load(data)


def read_query(path):
    fd = open(path, 'r')
    sqlFile = fd.read()
    fd.close()
    return sqlFile


def write_dotenv(env, file=".env"):
    """
    python-dotenv dont have any function to overwrite .env file
    write_dotenv is just a write variables in text file
    """
    with open(".env", "w") as f:
        for i in env:
            f.write(f'{i}={env[i]}\n')


def to_print_dict(dic: dict, spaces=4):
    """
    @ make the dict a key: value table string
    """
    ret = ''
    for i in dic:
        ret = ret + '\n' + f"{i}: {dic[i]}"
    return ret


def print_dict(dic: dict):
    """
    @ just print to_print_dict
    """
    print(to_print_dict(dic))


def sample(vec):
    """Return 1 random value of a vector 
    Args:
        vec :list: list or a pd.Series or pd.Dataframe
    Returns:
        what is inside the list/dataframe/series index []
    """
    return vec[randint(len(vec))]


def to_print_list(ls, skip_rows=False):
    """just remove [] from list to print it"""
    ret = ''
    skip = None
    if skip_rows:
        skip = '\n'
    for i in ls:
        ret = ret+f'{i},{skip}'
    return ret[:-1]


def print_list(ls):
    """print the list without []"""
    print(to_print_list(ls))



import pandas as pd
import numpy as np
from jh_utils.data.pandas.preprocessing import make_dummies


def create_time_series_dataframe(df,
                                 weekday_dummies=True,
                                 month_dummies=True,
                                 hour_dummies=True,
                                 datetime_column_name='datetime'):
    # shape = df.shape
    if weekday_dummies:
        df = pd.concat([df, make_dummies(
            pd.Series(df.iloc[:, 0].dt.weekday, name='weekday_dummie'))], axis=1)
    if month_dummies:
        df = pd.concat([df, make_dummies(
            pd.Series(df.iloc[:, 0].dt.month, name='month_dummie'))], axis=1)
    if hour_dummies:
        df = pd.concat([df, make_dummies(
            pd.Series(df.iloc[:, 0].dt.hour, name='hour_dummie'))], axis=1)

    df.index = df[datetime_column_name]
    df = df.iloc[:, 1:]
    return df


def get_number_hours_by_day(freq='1H'):
    freq = freq.upper()
    if freq == '1H':
        return 24
    if freq == '2H':
        return 12
    if freq == '3H':
        return 8
    if freq == '4H':
        return 6
    if freq == '6H':
        return 4
    if freq == '8H':
        return 3
    if freq == '12H':
        return 2


def create_sincos_year_array(n, start_date, end_date, n_days_in_year=366):
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)

    # day distance
    first_date = pd.to_datetime('{year}-01-01'.format(year=start_date.year))
    day_of_the_year_start = (start_date - first_date).days

    # year distance
    distance = (end_date - first_date).days

    ##
    start = (np.pi*2)*(day_of_the_year_start/n_days_in_year)
    end = np.pi*2*(distance/n_days_in_year)

    length = np.arange(start, end, (end-start)/n)
    df = pd.concat([pd.Series(np.sin(length)),
                   pd.Series(np.cos(length))], axis=1)
    df.columns = ['sin_year', 'cos_year']
    return df


def create_sincos_hour_array(hours_by_day, start_date, end_date):
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)
    # ! year distance
    distance = (end_date - start_date).days

    start = 0
    end = np.pi*2*distance
    length = np.arange(start, end, np.pi*2/(hours_by_day))
    df = pd.concat([pd.Series(np.sin(length)),
                   pd.Series(np.cos(length))], axis=1)
    df.columns = ['sin_day', 'cos_day']
    return df.iloc[:-1]



# importing libs
import pandas as pd
import json
from sqlalchemy import create_engine
import psycopg2


def get_top_rows(table, schema, engine, n=1):
    return get_sql_table(f'select * from {schema}.{table} dc limit {n}', engine)


def get_sql_table(query, engine, close_connection=False):
    """
    make a query in a database
    """
    # open connection
    engine.connect()

    # make select query
    table = pd.read_sql(query, engine)

    return table


def write_sql_table(
    table,
    table_name,
    schema,
    engine,
    if_exists='append',
    chunksize=10_000,
    index=False,
    close_connection=False,
):
    """
    if_exists => ["append","replace"]
    """

    # open connection
    engine.connect()

    # write or replace table
    table.to_sql(
        name=table_name,
        if_exists=if_exists,
        con=engine,
        schema=schema,
        method='multi',
        chunksize=chunksize,
        index=index,
    )

    # ## close connection
    if close_connection:
        engine.close()



import pandas as pd


def make_dummies(series):
    df = pd.get_dummies(series, drop_first=True)
    df.columns = list(map(lambda x: '{}_'.format(
        series.name) + str(x), list(df.columns)))
    return df



from sqlalchemy import inspect
import sqlalchemy as sa

# create_table_structure, get_tables, drop_table, delete_table, get_top_rows
# get_schemas, create_schema, drop_schema
# apply_delete_to_schema

# ! Table


def create_table_structure(
    pandas_df, table_name, engine, schema, index=False, if_exists='append'
):
    pandas_df.to_sql(
        name=table_name, con=engine, schema=schema, index=index, if_exists=if_exists
    )
    delete_table(table_name, schema, engine)


def get_tables(schema, engine):
    insp = inspect(engine)
    return insp.get_table_names(schema)


def drop_table(table, schema, engine, close_connection=True):
    conn = engine.connect()
    conn.execute(f'DROP TABLE IF EXISTS {schema}.{table};')
    if close_connection:
        conn.close()


def delete_table(table_name, schema, engine, close_connection=True):
    conn = engine.connect()
    conn.execute(f'DELETE FROM {schema}.{table_name}')
    if close_connection:
        conn.close()


# ! Schema


def get_schemas(engine):
    insp = inspect(engine)
    return insp.get_schema_names()


def create_schema(schema_name, engine):
    if not engine.dialect.has_schema(engine, schema_name):
        engine.execute(sa.schema.CreateSchema(schema_name))
    else:
        return 'Already exists'


def drop_schema(schema, engine, close_connection=True):
    conn = engine.connect()
    conn.execute(f'DROP SCHEMA {schema}')
    if close_connection:
        conn.close()


# ! delete all tables from a schema
def apply_delete_to_schema(schema, engine, close_connection=True):
    conn = engine.connect()
    tables = get_tables(schema, engine)
    for i in tables:
        delete_table(i, schema, engine)
    if close_connection:
        conn.close()



import pandas as pd
import ast


def create_dimension(dataframe, column_name):
    """
    [summary]

    Args:
        dataframe (dataframe): [description]
        column_name (string): [description]

    Returns:
        dataframe: [description]
    """

    dimension = dataframe[column_name]
    dimension = pd.DataFrame(dimension.unique()).reset_index()
    dimension.columns = ['id'+column_name, column_name]
    return dimension


def create_bridge_dimension(dataframe, column_name, index_column):
    temp_df = dataframe[[column_name, index_column]]
    temp_df.iloc[:, 0] = temp_df.iloc[:, 0].fillna(
        value='[]').apply(ast.literal_eval)

    # ! clean NA and explode the list
    temp_df = temp_df.explode(column_name)
    temp_df = temp_df[temp_df.iloc[:, 0].isna() == False]

    # ! create dimension
    dimension = create_dimension(temp_df, column_name)
    temp_df = temp_df.merge(dimension, on=column_name, how='left').iloc[:, 1:]
    return (temp_df, dimension)


def modelate_database(dataframe, columns=[], bridge_columns=[], index_column=None):
    ret = dict()

    # create the dimensions
    if columns != []:
        for i in columns:
            dim_temp = create_dimension(dataframe, i)
            ret[i] = dim_temp
            dataframe = dataframe.merge(dim_temp, on=i, how='left')

    if bridge_columns != []:
        if index_column == None:
            return "ERROR: Bridge tables need index column parameter"

        for i in bridge_columns:
            bridge_temp, dim_temp = create_bridge_dimension(
                dataframe, i, index_column)
            ret[i] = dim_temp
            ret['bridge_'+i] = bridge_temp

    # add fact to return
    ret['fact'] = dataframe.drop(columns + bridge_columns, axis=1)
    return ret



import pandas as pd
import numpy as np


def crossjoin(column_1, column_2):
    """
    Create a crossjoin dataframe based on 2 columns
    """
    columns_name = [column_1.name, column_2.name]
    column_1 = pd.DataFrame(column_1.unique())
    column_2 = pd.DataFrame(column_2.unique())
    column_1['key'], column_2['key'] = 0, 0
    df_ret = column_1.merge(column_2, on='key').drop(columns='key')
    df_ret.columns = columns_name
    return df_ret

##


def get_na_by_column(df):
    ret = df.isnull().mean()
    return ret


def clean_na(df, na_value=''):
    # Na from dataset

    df.index = df['date_time']
    df = df.drop(columns=['date_time'])

    df = df.replace(to_replace=na_value, value=np.NaN)
    df = df[df.iloc[:, 1].first_valid_index():]
    df = df.fillna(method='ffill')
    return df




def postgresql_snippet(table_name):
    query = f"""select column_name, data_type, character_maximum_length, column_default, is_nullable
    from INFORMATION_SCHEMA.COLUMNS where table_name = '{table_name}';"""
    return query



import pandas as pd
import numpy as np
from jh_utils.time_series.covariables import year_sincos_array, hour_sincos_array, hours_by_day
from jh_utils.data.pandas.preprocessing import make_dummies


def create_date_range_dataframe(start_date, end_date, freq='1H', column_name='hours'):
    df = pd.DataFrame({column_name: pd.date_range(
        start_date, end_date, freq=freq, closed='left')})
    return df


def create_time_series_dataframe(start_date, end_date, freq='1H', datetime_column_name='hours',
                                 weekday_dummies=True,
                                 month_dummies=True,
                                 hour_dummies=True):
    df = create_date_range_dataframe(start_date, end_date,
                                     freq=freq, column_name=datetime_column_name)
    if weekday_dummies:
        df = pd.concat([df, make_dummies(
            pd.Series(df.iloc[:, 0].dt.weekday, name='weekday_dummie'))], axis=1)
    if month_dummies:
        df = pd.concat([df, make_dummies(
            pd.Series(df.iloc[:, 0].dt.month, name='month_dummie'))], axis=1)
    if hour_dummies:
        df = pd.concat([df, make_dummies(
            pd.Series(df.iloc[:, 0].dt.hour, name='hour_dummie'))], axis=1)

    df.index = df[datetime_column_name]
    df = df.iloc[:, 1:]
    return df


def make_date_hour_crossjoin(column_1, column_2):
    """
    Create a crossjoin dataframe based on 2 columns
    """
    column_1 = pd.DataFrame(column_1.unique())
    column_2 = pd.DataFrame(column_2.unique())
    column_1['key'], column_2['key'] = 0, 0
    date_time = column_1.merge(column_2, on='key').drop(columns='key')
    date_time.columns = ['data', 'hora']
    date_time.sort_values(['data', 'hora'], inplace=True)
    date_time.reset_index(inplace=True, drop=True)
    return date_time



from datetime import datetime


class LOG():
    def __init__(self, log: dict) -> None:
        self.log = log
        self.log['log_created_at'] = datetime.now()

    def __repr__(self) -> str:
        return self.log



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


figsize = (20, 5)


def two_line_plot(df, x1, x2, freq='M', figsize=(20, 5), tittle=''):
    temp = df[['date_time', x1, x2]]
    temp.index = temp.date_time
    temp = temp.resample(freq).mean()
    temp.reset_index(inplace=True)
    plt.figure(figsize=figsize)
    _ = sns.lineplot(x=temp['date_time'], y=temp[x1]
                     ).set_title(tittle, fontsize=20)
    _ = sns.lineplot(x=temp['date_time'], y=temp[x2])


def two_hist_plot(df, x1, x2, freq='M', figsize=figsize, tittle=''):
    temp = df[['date_time', x1, x2]]
    temp.index = temp.date_time
    temp = temp.resample(freq).mean()
    temp.reset_index(inplace=True)
    plt.figure(figsize=figsize)
    _ = sns.histplot(df[x1], color='red').set_title(tittle, fontsize=20)
    _ = sns.histplot(df[x2])


def one_line_plot(df, x1, freq='M', figsize=figsize, tittle=''):
    temp = df[['date_time', x1]]
    temp.index = temp.date_time
    temp = temp.resample(freq).mean()
    temp.reset_index(inplace=True)
    plt.figure(figsize=figsize)
    _ = sns.lineplot(x=temp['date_time'], y=temp[x1]
                     ).set_title(tittle, fontsize=20)


def one_hist_plot(df, x1, freq='M', figsize=figsize, tittle=''):
    temp = df[['date_time', x1]]
    temp.index = temp.date_time
    temp = temp.resample(freq).mean()
    temp.reset_index(inplace=True)
    plt.figure(figsize=figsize)
    _ = sns.histplot(df[x1], color='red').set_title(tittle, fontsize=20)



import sqlalchemy as sa
import psycopg2


def create_string(user, password, host, db, port, sgbd="postgresql"):
    connection_string = sa.engine.URL.create(
        drivername=sgbd,
        username=user,
        password=password,
        host=host,
        database=db,
        port=port)
    return connection_string


def create_connection(database: str, user: str, password: str, host: str, port: str, sgbd='postgresql'):
    """
    create_connection(database=env['db'], 
                  user=env['user'],
                  password=env['pass'], 
                  host=env['host'],
                  port=env['port'])

    """
    con_string = create_string_connection(
        database, user, password, host, port, sgbd='postgresql')
    # Creating db string connection
    return sa.create_engine(con_string)


def create_string_connection(database: str, user: str, password: str, host: str, port: str, sgbd='postgresql'):
    """
    create_string_connection(database=env['db'], 
                  user=env['user'],
                  password=env['pass'], 
                  host=env['host'],
                  port=env['port'])
    """

    # Creating db string connection
    if sgbd == 'postgresql':
        con_string = f'postgresql://{user}:{password}@{host}:{port}/{database}'
    if sgbd == 'mysql':
        con_string = f'mysql+pymysql://{user}:{password}@{host}:{port}/{database}'
    if sgbd == 'redshift':
        con_string = f'redshift+psycopg2://{user}:{password}@{host}:{port}/{database}'
    return con_string
